# 
# Core functions for the Demand Prediction Challenge (dpc)
# 
# Contains functions for generating and fitting a model to data,
# plus some helper functions.

from dpcutil import *
from pylab import *
import datetime as dt
import json as js
import pandas as pd
import scipy.optimize as opt

# class to encapsulate a predictive model and its parameters
class Model:
    _G = None                    # Scale factor
    _F = None                    # Shape factor
    _z_G = None                  # Cached G params (currently unused)
    _z_F = [[1,1,1,1,1,1] for w in range(7)] # Cached F params (initial guess)
    
# Public methods

    # Initialize model by updating (i.e., fitting) with initial data
    def __init__ (self,h):
        self.update(h)
        return

    # fit the model with a new set of historical data
    def update (self,h):
        self._G = self._makeG(h)
        self._F = self._makeF(h/self._G(h.index))
        return

    # Evaluate the model at a set of points x
    def eval (self,x):
        return self._G(x)*self._F(x)

# Private methods

    # Make a function for the scale factor
    def _makeG (self,h):
        self.z_G = np.polyfit(self._hour(h.index),h,1)
        return lambda t: pd.Series((np.poly1d(self.z_G))(self._hour(t)), index=t)

    # Make a function for the shape factor
    def _makeF (self,h):
        f = lambda x,p0,p1,p2,p3,p4,p5: (np.poly1d([p0,p1,p2,p3,p4,p5]))(x)
        self._z_F = self._getSubModelParameters(h,f)
        return lambda t: pd.Series([f(k.hour,self._z_F[k.dayofweek][0],self._z_F[k.dayofweek][1],self._z_F[k.dayofweek][2],self._z_F[k.dayofweek][3],self._z_F[k.dayofweek][4],self._z_F[k.dayofweek][5]) for k in t], index=t)

    # Utility func to turn DatetimeIndex into float array (no side-effects)
    def _hour (self,x):
        return (x.asi8)/1e9/3600

    # Get fit params for shape factor sub-model (no side-effects)
    def _getSubModelParameters (self,x,f):
        daily = [x[(x.index.dayofweek==w)] for w in range(7)]
        return [opt.curve_fit(f,daily[d].index.hour,daily[d],self._z_F[d])[0] for d in range(7)]
